#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>
#include <type_traits>

// Define warp and block parameters
#define WARP_SIZE 32
#define WARPS_PER_BLOCK 8
#define RELU_BLOCK_SIZE 256

// Kernel to perform MLP forward pass with vectorized loads for improved memory coalescing.
// For float types, we use float4 (i.e., 4 floats per load); for double, we use double2 (2 doubles per load).

template <typename scalar_t>
__global__ void mlp_forward_kernel_coalesced(
    const scalar_t* __restrict__ input,
    const scalar_t* __restrict__ weight,
    const scalar_t* __restrict__ bias,
    scalar_t* __restrict__ output,
    int batch_size,
    int in_features,
    int out_features) {

    int lane = threadIdx.x;           // Lane index within a warp [0, 31]
    int warpId = threadIdx.y;           // Warp index within the block

    int batch_row = blockIdx.x;         // Each block in x-dim handles one batch row
    int neuron = blockIdx.y * WARPS_PER_BLOCK + warpId;  // Each warp computes one output neuron

    if (batch_row >= batch_size || neuron >= out_features) return;

    scalar_t sum = 0;

    // Determine vector width based on scalar size.
    // For float (4 bytes): use 4 (float4) -> 4 floats per load.
    // For double (8 bytes): use 2 (double2) -> 2 doubles per load.
    constexpr int vec_width = (sizeof(scalar_t) == 4 ? 4 : (sizeof(scalar_t) == 8 ? 2 : 1));

    int vec_count = in_features / vec_width;  // Number of complete vectorized loads
    int remainder = in_features % vec_width;  // Remaining elements

    if (vec_width > 1) {
        // Define vectorized type based on scalar_t
        using vec_t = typename std::conditional<sizeof(scalar_t) == 4, float4, double2>::type;

        // Reinterpret input pointers as vectorized pointers.
        const vec_t* input_vec = reinterpret_cast<const vec_t*>(input + batch_row * in_features);
        const vec_t* weight_vec = reinterpret_cast<const vec_t*>(weight + neuron * in_features);

        // Each thread in the warp processes multiple vectorized chunks with stride equal to warp size
        for (int idx = lane; idx < vec_count; idx += WARP_SIZE) {
            vec_t in_val = input_vec[idx];
            vec_t w_val = weight_vec[idx];
            if constexpr (sizeof(scalar_t) == 4) {
                sum += in_val.x * w_val.x + in_val.y * w_val.y + in_val.z * w_val.z + in_val.w * w_val.w;
            } else if constexpr (sizeof(scalar_t) == 8) {
                sum += in_val.x * w_val.x + in_val.y * w_val.y;
            }
        }
    } else {
        // Fallback to scalar loads if vectorization is not applicable
        for (int i = lane; i < in_features; i += WARP_SIZE) {
            sum += input[batch_row * in_features + i] * weight[neuron * in_features + i];
        }
    }

    // Handle remainder elements if any
    int start = vec_count * vec_width;
    for (int i = start + lane; i < in_features; i += WARP_SIZE) {
        sum += input[batch_row * in_features + i] * weight[neuron * in_features + i];
    }

    // Warp-level reduction using shuffle operations
    for (int offset = WARP_SIZE / 2; offset > 0; offset /= 2) {
        sum += __shfl_down_sync(0xffffffff, sum, offset);
    }

    // The first thread in the warp writes the result along with bias
    if (lane == 0) {
        output[batch_row * out_features + neuron] = sum + bias[neuron];
    }
}

// ReLU kernel (unchanged, with a stride loop for full coverage)
template <typename scalar_t>
__global__ void relu_kernel_coalesced(scalar_t* __restrict__ data, int size) {
    int idx = blockIdx.x * blockDim.x + threadIdx.x;
    int stride = blockDim.x * gridDim.x;
    for (; idx < size; idx += stride) {
        data[idx] = data[idx] > 0 ? data[idx] : static_cast<scalar_t>(0);
    }
}

// Main forward function applying the MLP layer by layer

torch::Tensor mlp_cuda_forward(
    torch::Tensor input,
    std::vector<torch::Tensor> weights,
    std::vector<torch::Tensor> biases) {

    auto device = input.device();
    auto num_layers = weights.size();
    torch::Tensor current = input;

    for (size_t layer = 0; layer < num_layers; layer++) {
        const int batch_size = current.size(0);
        const int in_features = current.size(1);
        const int out_features = weights[layer].size(0);

        auto output = torch::empty({batch_size, out_features}, 
                                     torch::dtype(current.dtype()).device(device));

        // Configure grid and block dimensions
        // Block: 32 threads in x (warp size) and WARPS_PER_BLOCK in y
        dim3 block(WARP_SIZE, WARPS_PER_BLOCK);
        dim3 grid(batch_size, (out_features + WARPS_PER_BLOCK - 1) / WARPS_PER_BLOCK);

        AT_DISPATCH_FLOATING_TYPES(current.scalar_type(), "mlp_forward_kernel_coalesced", ([&] {
            mlp_forward_kernel_coalesced<scalar_t><<<grid, block>>>(
                current.data_ptr<scalar_t>(),
                weights[layer].data_ptr<scalar_t>(),
                biases[layer].data_ptr<scalar_t>(),
                output.data_ptr<scalar_t>(),
                batch_size,
                in_features,
                out_features
            );
        }));

        // Apply ReLU activation for all layers except the final one
        if (layer < num_layers - 1) {
            int size = batch_size * out_features;
            int num_blocks = (size + RELU_BLOCK_SIZE - 1) / RELU_BLOCK_SIZE;
            AT_DISPATCH_FLOATING_TYPES(output.scalar_type(), "relu_kernel_coalesced", ([&] {
                relu_kernel_coalesced<scalar_t><<<num_blocks, RELU_BLOCK_SIZE>>>(
                    output.data_ptr<scalar_t>(),
                    size
                );
            }));
        }

        current = output;
    }

    return current;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
  m.def("forward", &mlp_cuda_forward, "MLP forward (CUDA coalesced vectorized)");
}
